
;;;;; (c) Remi van Trijp
;;;;;
;;;;; This file contains render and de-render functions tailored for the demo grammar.
;;;;; -----------------------------------------------------------------------------------------------------

(in-package :ldd-demo)

(defmethod create-initial-structure ((meaning list) (mode (eql :root-node-mode)))
  (make-instance 'coupled-feature-structure
		 :left-pole `((root
			       (referent)
			       (meaning ,meaning)
			       (sem-cat nil)))
		 :right-pole '((root
				(form nil)
				(syn-cat nil)))
		 :left-pole-domain 'sem
		 :right-pole-domain 'syn))

(defmethod de-render ((utterance list) (mode (eql :de-render-with-sequence-and-precedes)))
  (let ((strings nil)
        (sequence nil)
	(constraints nil))
    (dolist (string utterance)
      (let ((new (make-const string nil)))
        (push new sequence)
	(dolist (prev strings)
	  (push `(precedes ,(second prev) ,new) constraints))
	(push `(string ,new ,string) strings)))
    (do ((left strings (rest left)))
	((null (cdr left)))
      (push `(meets ,(second (second left)) ,(second (first left)))
	    constraints))
    (make-instance 'coupled-feature-structure 
		   :left-pole '((root (meaning ()) (sem-cat ())))
		   :right-pole `((root (form ,(cons (cons 'sequence (reverse sequence))
                                                   (append strings constraints)))
				      (syn-cat ()))))))

(defun insert-precede (order lst)
  (cond
   ((null lst) order) ;; We reached the end of the list, so just insert the units.
   ((eql (first lst) (first order)) ;; The first unit is already found early in the list.
    (if (member (second order) lst) ;; Do not change anything if the second unit is already there
      lst
      (append order (rest lst)))) ;; If not, add the second unit as well.
   ((eql (first lst) (second order)) ;; If the second unit is found, insert the first unit here.
    (cons (first order) (remove (first order) lst)))
   (t
    (cons (first lst) (insert-precede order (rest lst))))))

(defun insert-precedes (order chains)
  "Check whether all the precedes constraints are also satisfied."
  (cond
   ((null chains) (list order))
   ((find order chains :test #'subsetp) chains)
   ((member (first order) (first chains))
    (if (find (second order) (rest chains) :test #'member)
      chains
      (append (list (first chains) (list (second order))) (rest chains))))
   ((member (second order) (first chains))
    (let ((preceding-chain (find (first order) (rest chains) :test #'member)))
      (if preceding-chain
        (append (list preceding-chain (first chains))
                (remove preceding-chain (rest chains) :test #'equal))
        (cons (list (first order)) chains))))
   (t
    (cons (first chains) (insert-precedes order (rest chains))))))

(defun insert-meets (meets-lst lst)
  "Take care of the adjacency constraints."
  (if (null lst) ;; If the list is currently empty...
    (insert-meets meets-lst (flatten meets-lst)) ;; Then order the 
    (let* ((meets (assoc (first lst) meets-lst))
           (next-unit (second meets)))
      (if next-unit
        (cons (first lst)
              (insert-meets (remove meets meets-lst :test #'equal)
                            (cons next-unit (rest lst))))
        (cons (first lst) (insert-meets meets-lst (rest lst)))))))

(defun insert-meets-constraint (meeting-units lst)
  "Insert the meets-constraint in its proper position."
  (cond
   ((null lst) (list meeting-units))
   ((member meeting-units lst :test #'equal) lst)
   ((eql (first meeting-units) (second (first lst)))
    (sort-order-constraints (rest lst) ;; Dangerous recursion...
                            (list (first lst) meeting-units)))
   ((eql (second meeting-units) (first (first lst)))
    (sort-order-constraints lst (list meeting-units)))
   (t
    (cons (first lst) (insert-meets-constraint meeting-units (rest lst))))))

(defun sort-order-constraints (meets-constraints &optional solution)
  "Make sure that the meets-constraints are in proper order."
  (if (null meets-constraints)
    solution
    (sort-order-constraints (rest meets-constraints)
                            (insert-meets-constraint (first meets-constraints) solution))))

(defun make-chain (lst-1 lst-2)
  "Merge the two lists by appending list-1 to the rest of lst-2."
  (append lst-1 (rest lst-2)))

(defun make-chains (meeting-units)
  "If there are chains of meets-constraints, group them as one list."
  (when meeting-units
    (let* ((meets-1 (first meeting-units))
           (meets-2 (assoc (first (last meets-1)) meeting-units)))
      (if meets-2
        (make-chains (cons (make-chain meets-1 meets-2)
                           (remove meets-2 (rest meeting-units) :test #'equal)))
        (cons (first meeting-units) (make-chains (rest meeting-units)))))))

(defun enforce-first-unit (unit chains)
  "Pick out the chain that starts with the initial unit and put it in front."
  (let ((chain (assoc unit chains)))
    (if chain
      (cons chain (remove chain chains :test #'equal))
      (cons (list unit) chains))))

(defun flatten-sequence (chains)
  "Return a flat list of unique unit-names."
  (let (result)
    (dolist (chain chains)
      (if (symbolp chain)
        (pushnew chain result)
        (dolist (unit-name chain)
          (pushnew unit-name result))))
    (reverse result)))

(defun expand-domains-in-meets (meets-constraints units-and-domains)
  (when meets-constraints
    (let* ((constraint (first meets-constraints))
           (first-unit-domain (alist-value (first constraint) units-and-domains))
           (second-unit-domain (alist-value (second constraint) units-and-domains)))
      (cond
       (first-unit-domain (expand-domains-in-meets
                           (cons (list (last-elt first-unit-domain) (second constraint))
                                 (rest meets-constraints))
                           units-and-domains))
       (second-unit-domain (expand-domains-in-meets
                            (cons (list (first constraint) (first second-unit-domain))
                                  (rest meets-constraints))
                            units-and-domains))
       (t
        (cons constraint (expand-domains-in-meets (rest meets-constraints) units-and-domains)))))))

(defun expand-domains-in-precedes (precedes-constraints units-and-domains)
  (when precedes-constraints
    (let* ((constraint (first precedes-constraints))
           (first-unit-domain (alist-value (first constraint) units-and-domains))
           (second-unit-domain (alist-value (second constraint) units-and-domains)))
      (cond (first-unit-domain (expand-domains-in-precedes
                                (append (loop for unit-name in first-unit-domain
                                              with other-unit = (second constraint)
                                              collect `(,unit-name ,other-unit))
                                        (rest precedes-constraints))
                                units-and-domains))
            (second-unit-domain (expand-domains-in-precedes
                                 (append (loop for unit-name in second-unit-domain
                                               with other-unit = (first constraint)
                                               collect `(,other-unit ,unit-name))
                                         (rest precedes-constraints))
                                 units-and-domains))
            (t
             (cons (first precedes-constraints)
                   (expand-domains-in-precedes (rest precedes-constraints) units-and-domains)))))))

(defun render-with-domains (struct)
  (let (sequence strings meets precedes first-unit units-and-domains)
    (dolist (unit struct)
      (let ((form (unit-feature-value unit 'form)))
        (dolist (value form)
          (let ((f (first value)))
            (cond
             ((eql f 'string) (push (rest value) strings))
             ((eql f 'meets) (push (rest value) meets))
             ((eql f 'precedes) (push (rest value) precedes))
             ((eql f 'sequence) (setf first-unit (second value)))
             ((eql f 'domain) (push (list (unit-name unit) (flatten (second value))) units-and-domains))
             (t
              nil))))))
    (setf meets (pp (remove-duplicates (expand-domains-in-meets meets units-and-domains) :test #'equal)))
    (setf precedes (pp (remove-duplicates (expand-domains-in-precedes precedes units-and-domains) :test #'equal)))
    (setf sequence (sort-order-constraints meets))
    (setf sequence (make-chains sequence))
    (when first-unit
      (setf sequence (enforce-first-unit first-unit sequence)))
    (dolist (order (sort-order-constraints precedes))
      (setf sequence (insert-precedes order sequence)))
    (let* ((flat-sequence (flatten-sequence sequence))
           (ordered-forms (loop for unit in flat-sequence
                                for f = (assoc unit strings)
                                when f
                                collect (second f)))
           (other-strings (loop for unit in strings
                                unless (member (first unit) flat-sequence)
                                collect (second unit))))
      (append ordered-forms other-strings))))

(defun my-render (syn)
  "Tailored for this English grammar."
  ;;(setf *syn* syn)
  (let (sequence strings meets precedes first-unit)
    (dolist (unit syn)
      (let ((form (unit-feature-value unit 'form)))
        (dolist (value form)
          (let ((f (first value)))
          (cond
           ((eql f 'string) (push (rest value) strings))
           ((eql f 'meets) (push (rest value) meets))
           ((eql f 'precedes) (push (rest value) precedes))
           ((eql f 'sequence) (setf first-unit (second value)))
           (t
            nil))))))
    (setf sequence (sort-order-constraints meets))
    (setf sequence (make-chains sequence))
    (when first-unit
      (setf sequence (enforce-first-unit first-unit sequence)))
    (dolist (order (sort-order-constraints precedes))
      (setf sequence (insert-precedes order sequence)))
    (let* ((flat-sequence (flatten-sequence sequence))
           (ordered-forms (loop for unit in flat-sequence
                                for f = (assoc unit strings)
                                when f
                                collect (second f)))
           (other-strings (loop for unit in strings
                                unless (member (first unit) flat-sequence)
                                collect (second unit))))
      (append ordered-forms other-strings))))

(defmethod render ((cfs coupled-feature-structure) (mode (eql :my-render)))
  (my-render (right-pole-structure cfs)))

(defmethod render ((struct list)  (mode (eql :my-render)))
  (my-render struct))

(defmethod render ((cfs coupled-feature-structure) (mode (eql :use-domains)))
  (render-with-domains (right-pole-structure cfs)))

(defmethod render ((struct list) (mode (eql :use-domains)))
  (render-with-domains struct))
