
;;;;; (c) Remi van Trijp

(in-package :ldd-demo)

;;;;; Useful macros.
;;;;; ----------------------------------------------------------------------------------------
(export '(print-expansion when-let when-let* while last-elt cond-let))

(defmacro print-expansion (macro-call)
  "Pretty print the macro-expansion of a macro-call."
  `(pprint (macroexpand-1 ,macro-call)))

(defmacro when-let* (vars-and-tests then-clause &optional else-clause)
  "If a value can be found for all vars, then execute the then-clause."
  `(let* ,vars-and-tests
     (if ,(cons 'and (mapcar #'first vars-and-tests))
       ,then-clause
       ,else-clause)))

(defmacro cond-let (&body body)
  (if (null body)
    nil
    `(let ,(list (caar body))
       (if ,(caaar body)
         (progn ,@(cdar body))
         (cond-let ,@(rest body))))))
        
(defmacro when-let (vars-and-tests then-clause &optional else-clause)
  "If a value can be found for all vars, then execute the then-clause."
  `(let ,vars-and-tests
     (if ,(cons 'and (mapcar #'first vars-and-tests))
       ,then-clause
       ,else-clause)))

(defmacro while (test &body body)
  "Easier to read than do macro."
  `(do ()
       ((not ,test))
     ,@body))

;;;;; Some helper functions.
;;;;; ------------------------------------------------------------------------------
(defun bigger-list? (lst1 lst2)
  "More efficient function than comparing lengths of lists, capable of handling cons-lists."
  (cond
   ((null lst1) nil)
   ((null lst2) t)
   (t
    (bigger-list? (listify (rest lst1)) (listify (rest lst2))))))

(defun load-my-file (filename &optional (directory '(:relative "examples" "English-grammar")))
  (load (merge-pathnames (make-pathname :directory directory
                                        :name filename
                                        :type "lisp")
                         (babel-pathname))))

(proclaim '(inline get-cxn make-cxn-name show-cxn pp-cxn transient-structure
                   syn-pole sem-pole root-unit unit-body last-elt cons-if >1))

(defun >1 (lst)
  "Checks whether a list is bigger than 1."
  (rest lst))

(defun first-atom (lst)
  "Recursively fetch the first atom of a list."
  (let ((x (first lst)))
    (if (atom x)
      x
      (first-atom x))))

(defun last-atom (lst)
  "recursively fetch the last atom in a list."
  (let ((x (last lst)))
    (cond ((rest x) (last-atom x))
          ((atom (first x)) (first x))
          (t
           (last-atom (first x))))))

(defun cons-if (x lst)
  "If x, then cons x in lst."
  (if x (cons x lst) lst))

(defun alist-value (elt alist &key (key #'second))
  "Returns the second element of an alist if found."
  (funcall key (assoc elt alist)))

(defun last-elt (lst)
  "Fetch the last element of a list."
  (first (last lst)))

(defun get-cxn (construction-name)
  "Find a construction in *constructions*."
  (find-cxn construction-name *constructions*))

(defun make-cxn-name (id &optional (type 'cxn))
  "Creates a name for a construction."
  (make-symbol (format nil "~a-~a" id type)))

(defun show-cxn (construction-name)
  "Show the construction in the web browser."
  (show (get-cxn construction-name)))

(defun pp-cxn (construction-name)
  "Pretty-print a construction."
  (pp (get-cxn construction-name)))

(defun transient-structure (cip)
  "Returns the transient structure of a construction-inventory-processor."
  (car-resulting-cfs (cipn-car cip)))

(defun syn-pole (transient-structure)
  "Return the syntactic pole of a transient structure."
  (right-pole-structure transient-structure))

(defun sem-pole (transient-structure)
  "Return the semantic pole of a transient structure."
  (left-pole-structure transient-structure))

(defun root-unit (pole)
  "Return the root-unit of a pole."
  (assoc 'root-unit pole))

(defun unit-body (unit)
  "Returns the unit without the unit-name."
  (rest unit))

(defun replace-feature-value-in-unit (new-value old-value feature-name unit)
  "Replace the value of a feature with a new value within a unit."
  (if (assoc feature-name (unit-body unit))
    (cons (unit-name unit)
          (subst (list feature-name new-value)
                 (list feature-name old-value)
                 (unit-body unit)
                 :test #'equal))
    (cons (unit-name unit)
          (cons (list feature-name new-value) (unit-body unit)))))

(defun replace-unit (new-unit old-unit pole)
  "Replace a unit in a pole."
  (substitute new-unit old-unit pole :test #'equal))
